@@ -35,7 +35,7 @@
 
 #include "videodev2.h"
 
-#include "ivtv.h"
+#include "media/ivtv.h"
 
 int main(int argc, char **argv)
 {
