@@ -1,8 +1,9 @@
 /*
     Copyright (C) 2003-2004  Kevin Thayer <nufan_wfk at yahoo dot com>
 
-    Cleanup and VBI and audio in/out options:
-    Copyright (C) 2004  Hans Verkuil <hverkuil@xs4all.nl>
+    Cleanup and VBI and audio in/out options, introduction in v4l-dvb,
+    support for most new APIs since 2006.
+    Copyright (C) 2004, 2006, 2007  Hans Verkuil <hverkuil@xs4all.nl>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -79,6 +80,8 @@ enum Option {
 	OptSetSlicedVbiOutFormat,
 	OptGetOverlayFormat,
 	//OptSetOverlayFormat, TODO
+	OptGetOutputOverlayFormat,
+	OptSetOutputOverlayFormat,
 	OptGetVbiFormat,
 	//OptSetVbiFormat, TODO
 	OptGetVbiOutFormat,
@@ -94,19 +97,33 @@ enum Option {
 	OptSetVideoOutFormat,
 	OptGetSlicedVbiCap,
 	OptGetSlicedVbiOutCap,
-	OptGetVideoCrop,
-	OptSetVideoCrop,
+	OptGetFBuf,
+	OptSetFBuf,
+	OptGetCrop,
+	OptSetCrop,
+	OptGetOutputCrop,
+	OptSetOutputCrop,
+	OptGetOverlayCrop,
+	OptSetOverlayCrop,
+	OptGetOutputOverlayCrop,
+	OptSetOutputOverlayCrop,
 	OptGetAudioInput,
 	OptSetAudioInput,
 	OptGetAudioOutput,
 	OptSetAudioOutput,
 	OptListAudioOutputs,
 	OptListAudioInputs,
+	OptGetCropCap,
+	OptGetOutputCropCap,
+	OptGetOverlayCropCap,
+	OptGetOutputOverlayCropCap,
 	OptLast = 256
 };
 
 static char options[OptLast];
 
+static unsigned capabilities;
+
 typedef std::vector<struct v4l2_ext_control> ctrl_list;
 static ctrl_list user_ctrls;
 static ctrl_list mpeg_ctrls;
@@ -136,8 +153,10 @@ static const flag_def service_def[] = {
 };
 
 /* fmts specified */
-#define FMTWidth		(1L<<0)
-#define FMTHeight		(1L<<1)
+#define FmtWidth		(1L<<0)
+#define FmtHeight		(1L<<1)
+#define FmtChromaKey		(1L<<2)
+#define FmtGlobalAlpha		(1L<<3)
 
 /* crop specified */
 #define CropWidth		(1L<<0)
@@ -183,6 +202,8 @@ static struct option long_options[] = {
 	{"verbose", no_argument, 0, OptVerbose},
 	{"log-status", no_argument, 0, OptLogStatus},
 	{"get-fmt-overlay", no_argument, 0, OptGetOverlayFormat},
+	{"get-fmt-output-overlay", no_argument, 0, OptGetOutputOverlayFormat},
+	{"set-fmt-output-overlay", required_argument, 0, OptSetOutputOverlayFormat},
 	{"get-fmt-sliced-vbi", no_argument, 0, OptGetSlicedVbiFormat},
 	{"set-fmt-sliced-vbi", required_argument, 0, OptSetSlicedVbiFormat},
 	{"get-fmt-sliced-vbi-out", no_argument, 0, OptGetSlicedVbiOutFormat},
@@ -191,8 +212,20 @@ static struct option long_options[] = {
 	{"get-fmt-vbi-out", no_argument, 0, OptGetVbiOutFormat},
 	{"get-sliced-vbi-cap", no_argument, 0, OptGetSlicedVbiCap},
 	{"get-sliced-vbi-out-cap", no_argument, 0, OptGetSlicedVbiOutCap},
-	{"get-crop-video", no_argument, 0, OptGetVideoCrop},
-	{"set-crop-video", required_argument, 0, OptSetVideoCrop},
+	{"get-fbuf", no_argument, 0, OptGetFBuf},
+	{"set-fbuf", required_argument, 0, OptSetFBuf},
+	{"get-cropcap", no_argument, 0, OptGetCropCap},
+	{"get-crop", no_argument, 0, OptGetCrop},
+	{"set-crop", required_argument, 0, OptSetCrop},
+	{"get-cropcap-output", no_argument, 0, OptGetOutputCropCap},
+	{"get-crop-output", no_argument, 0, OptGetOutputCrop},
+	{"set-crop-output", required_argument, 0, OptSetOutputCrop},
+	{"get-cropcap-overlay", no_argument, 0, OptGetOverlayCropCap},
+	{"get-crop-overlay", no_argument, 0, OptGetOverlayCrop},
+	{"set-crop-overlay", required_argument, 0, OptSetOverlayCrop},
+	{"get-cropcap-output-overlay", no_argument, 0, OptGetOutputOverlayCropCap},
+	{"get-crop-output-overlay", no_argument, 0, OptGetOutputOverlayCrop},
+	{"set-crop-output-overlay", required_argument, 0, OptSetOutputOverlayCrop},
 	{0, 0, 0, 0}
 };
 
@@ -253,17 +286,21 @@ static void usage(void)
 	printf("  --list-formats     display supported video formats [VIDIOC_ENUM_FMT]\n");
 	printf("  -V, --get-fmt-video\n");
 	printf("     		     query the video capture format [VIDIOC_G_FMT]\n");
-	printf("  -v, --set-fmt-video=width=<x>,height=<y>\n");
+	printf("  -v, --set-fmt-video=width=<w>,height=<h>\n");
 	printf("                     set the video capture format [VIDIOC_S_FMT]\n");
 	printf("  --verbose          turn on verbose ioctl error reporting.\n");
 	printf("\n");
 	printf("Uncommon options:\n");
 	printf("  --get-fmt-video-out\n");
 	printf("     		     query the video output format [VIDIOC_G_FMT]\n");
-	printf("  --set-fmt-video-out=width=<x>,height=<y>\n");
+	printf("  --set-fmt-video-out=width=<w>,height=<h>\n");
 	printf("                     set the video output format [VIDIOC_S_FMT]\n");
 	printf("  --get-fmt-overlay\n");
 	printf("     		     query the video overlay format [VIDIOC_G_FMT]\n");
+	printf("  --get-fmt-output-overlay\n");
+	printf("     		     query the video output overlay format [VIDIOC_G_FMT]\n");
+	printf("  --set-fmt-output-overlay=chromakey=<key>,global_alpha=<alpha>\n");
+	printf("     		     set the video output overlay format [VIDIOC_S_FMT]\n");
 	printf("  --get-sliced-vbi-cap\n");
 	printf("		     query the sliced VBI capture capabilities [VIDIOC_G_SLICED_VBI_CAP]\n");
 	printf("  --get-sliced-vbi-out-cap\n");
@@ -280,10 +317,30 @@ static void usage(void)
 	printf("                     vps:      VPS (PAL/SECAM)\n");
 	printf("  --get-fmt-vbi      query the VBI capture format [VIDIOC_G_FMT]\n");
 	printf("  --get-fmt-vbi-out  query the VBI output format [VIDIOC_G_FMT]\n");
-	printf("  --get-crop-video\n");
-	printf("     		     query the video capture crop window [VIDIOC_G_CROP]\n");
-	printf("  --set-crop-video=top=<x>,left=<y>,width=<w>,height=<h>\n");
+	printf("  --get-fbuf         query the overlay framebuffer data [VIDIOC_G_FBUF]\n");
+	printf("  --set-fbuf=chromakey=<0/1>,global_alpha=<0/1>,local_alpha=<0/1>,\n");
+	printf("             global_inv_alpha=<0/1>,local_inv_alpha=<0/1>\n");
+	printf("		     set the overlay framebuffer [VIDIOC_S_FBUF]\n");
+	printf("  --get-cropcap      query the crop capabilities [VIDIOC_CROPCAP]\n");
+	printf("  --get-crop	     query the video capture crop window [VIDIOC_G_CROP]\n");
+	printf("  --set-crop=top=<x>,left=<y>,width=<w>,height=<h>\n");
 	printf("                     set the video capture crop window [VIDIOC_S_CROP]\n");
+	printf("  --get-cropcap-output\n");
+	printf("                     query the crop capabilities for video output [VIDIOC_CROPCAP]\n");
+	printf("  --get-crop-output  query the video output crop window [VIDIOC_G_CROP]\n");
+	printf("  --set-crop-output=top=<x>,left=<y>,width=<w>,height=<h>\n");
+	printf("                     set the video output crop window [VIDIOC_S_CROP]\n");
+	printf("  --get-cropcap-overlay\n");
+	printf("                     query the crop capabilities for video overlay [VIDIOC_CROPCAP]\n");
+	printf("  --get-crop-overlay query the video overlay crop window [VIDIOC_G_CROP]\n");
+	printf("  --set-crop-overlay=top=<x>,left=<y>,width=<w>,height=<h>\n");
+	printf("                     set the video overlay crop window [VIDIOC_S_CROP]\n");
+	printf("  --get-cropcap-output-overlay\n");
+	printf("                     query the crop capabilities for video output overlays [VIDIOC_CROPCAP]\n");
+	printf("  --get-crop-output-overlay\n");
+	printf("                     query the video output overlay crop window [VIDIOC_G_CROP]\n");
+	printf("  --set-crop-output-overlay=top=<x>,left=<y>,width=<w>,height=<h>\n");
+	printf("                     set the video output overlay crop window [VIDIOC_S_CROP]\n");
 	printf("  --get-audio-input  query the audio input [VIDIOC_G_AUDIO]\n");
 	printf("  --set-audio-input=<num>\n");
 	printf("                     set the audio input to <num> [VIDIOC_S_AUDIO]\n");
@@ -327,6 +384,8 @@ static std::string buftype2s(int type)
 		return "Sliced VBI Capture";
 	case V4L2_BUF_TYPE_SLICED_VBI_OUTPUT:
 		return "Sliced VBI Output";
+	case V4L2_BUF_TYPE_VIDEO_OUTPUT_OVERLAY:
+		return "Video Output Overlay";
 	case V4L2_BUF_TYPE_PRIVATE:
 		return "Private";
 	default:
@@ -364,6 +423,10 @@ static std::string field2s(int val)
 		return "Sequential Bottom-Top";
 	case V4L2_FIELD_ALTERNATE:
 		return "Alternating";
+	case V4L2_FIELD_INTERLACED_TB:
+		return "Interlaced Top-Bottom";
+	case V4L2_FIELD_INTERLACED_BT:
+		return "Interlaced Bottom-Top";
 	default:
 		return "Unknown (" + num2s(val) + ")";
 	}
@@ -409,7 +472,7 @@ static std::string flags2s(unsigned val, const flag_def *def)
 
 static void print_sliced_vbi_cap(struct v4l2_sliced_vbi_cap &cap)
 {
-//	printf("\tType           : %s\n", buftype2s(vfmt.type).c_str());
+	printf("\tType           : %s\n", buftype2s(cap.type).c_str());
 	printf("\tService Set    : %s\n",
 			flags2s(cap.service_set, service_def).c_str());
 	for (int i = 0; i < 24; i++) {
@@ -575,12 +638,89 @@ static void find_controls(int fd)
 	}
 }
 
+static std::string fbufcap2s(unsigned cap)
+{
+	std::string s;
+
+	if (cap & V4L2_FBUF_CAP_EXTERNOVERLAY)
+		s += "\t\t\tExtern Overlay\n";
+	if (cap & V4L2_FBUF_CAP_CHROMAKEY)
+		s += "\t\t\tChromakey\n";
+	if (cap & V4L2_FBUF_CAP_GLOBAL_ALPHA)
+		s += "\t\t\tGlobal Alpha\n";
+	if (cap & V4L2_FBUF_CAP_LOCAL_ALPHA)
+		s += "\t\t\tLocal Alpha\n";
+	if (cap & V4L2_FBUF_CAP_GLOBAL_INV_ALPHA)
+		s += "\t\t\tGlobal Inverted Alpha\n";
+	if (cap & V4L2_FBUF_CAP_LOCAL_INV_ALPHA)
+		s += "\t\t\tLocal Inverted Alpha\n";
+	if (cap & V4L2_FBUF_CAP_LIST_CLIPPING)
+		s += "\t\t\tClipping List\n";
+	if (cap & V4L2_FBUF_CAP_BITMAP_CLIPPING)
+		s += "\t\t\tClipping Bitmap\n";
+	if (s.empty()) s += "\t\t\t\n";
+	return s;
+}
+
+static std::string fbufflags2s(unsigned fl)
+{
+	std::string s;
+
+	if (fl & V4L2_FBUF_FLAG_PRIMARY)
+		s += "\t\t\tPrimary Graphics Surface\n";
+	if (fl & V4L2_FBUF_FLAG_OVERLAY)
+		s += "\t\t\tOverlay Matches Capture/Output Size\n";
+	if (fl & V4L2_FBUF_FLAG_CHROMAKEY)
+		s += "\t\t\tChromakey\n";
+	if (fl & V4L2_FBUF_FLAG_GLOBAL_ALPHA)
+		s += "\t\t\tGlobal Alpha\n";
+	if (fl & V4L2_FBUF_FLAG_LOCAL_ALPHA)
+		s += "\t\t\tLocal Alpha\n";
+	if (fl & V4L2_FBUF_FLAG_GLOBAL_INV_ALPHA)
+		s += "\t\t\tGlobal Inverted Alpha\n";
+	if (fl & V4L2_FBUF_FLAG_LOCAL_INV_ALPHA)
+		s += "\t\t\tLocal Inverted Alpha\n";
+	if (s.empty()) s += "\t\t\t\n";
+	return s;
+}
+
+static void printfbuf(const struct v4l2_framebuffer &fb)
+{
+	int is_ext = fb.capability & V4L2_FBUF_CAP_EXTERNOVERLAY;
+
+	printf("Framebuffer Format:\n");
+	printf("\tCapability    : %s", fbufcap2s(fb.capability).c_str() + 3);
+	printf("\tFlags         : %s", fbufflags2s(fb.flags).c_str() + 3);
+	if (fb.base)
+		printf("\tBase          : 0x%p\n", fb.base);
+	printf("\tWidth         : %d\n", fb.fmt.width);
+	printf("\tHeight        : %d\n", fb.fmt.height);
+	printf("\tPixel Format  : %s\n", fcc2s(fb.fmt.pixelformat).c_str());
+	if (!is_ext) {
+		printf("\tBytes per Line: %d\n", fb.fmt.bytesperline);
+		printf("\tSize image    : %d\n", fb.fmt.sizeimage);
+		printf("\tColorspace    : %s\n", colorspace2s(fb.fmt.colorspace).c_str());
+		if (fb.fmt.priv)
+			printf("\tCustom Info   : %08x\n", fb.fmt.priv);
+	}
+}
+
 static void printcrop(const struct v4l2_crop &crop)
 {
 	printf("Crop: Left %d, Top %d, Width %d, Height %d\n",
 			crop.c.left, crop.c.top, crop.c.width, crop.c.height);
 }
 
+static void printcropcap(const struct v4l2_cropcap &cropcap)
+{
+	printf("Crop Capability %s:\n", buftype2s(cropcap.type).c_str());
+	printf("\tBounds      : Left %d, Top %d, Width %d, Height %d\n",
+			cropcap.bounds.left, cropcap.bounds.top, cropcap.bounds.width, cropcap.bounds.height);
+	printf("\tDefault     : Left %d, Top %d, Width %d, Height %d\n",
+			cropcap.defrect.left, cropcap.defrect.top, cropcap.defrect.width, cropcap.defrect.height);
+	printf("\tPixel Aspect: %u/%u\n", cropcap.pixelaspect.numerator, cropcap.pixelaspect.denominator);
+}
+
 static int printfmt(struct v4l2_format vfmt)
 {
 	const flag_def vbi_def[] = {
@@ -588,12 +728,11 @@ static int printfmt(struct v4l2_format vfmt)
 		{ V4L2_VBI_INTERLACED, "interlaced" },
 		{ 0, NULL }
 	};
-	printf("Format:\n");
+	printf("Format %s:\n", buftype2s(vfmt.type).c_str());
 
 	switch (vfmt.type) {
 	case V4L2_BUF_TYPE_VIDEO_CAPTURE:
 	case V4L2_BUF_TYPE_VIDEO_OUTPUT:
-		printf("\tType          : %s\n", buftype2s(vfmt.type).c_str());
 		printf("\tWidth/Height  : %u/%u\n", vfmt.fmt.pix.width, vfmt.fmt.pix.height);
 		printf("\tPixel Format  : %s\n", fcc2s(vfmt.fmt.pix.pixelformat).c_str());
 		printf("\tField         : %s\n", field2s(vfmt.fmt.pix.field).c_str());
@@ -603,20 +742,20 @@ static int printfmt(struct v4l2_format vfmt)
 		if (vfmt.fmt.pix.priv)
 			printf("\tCustom Info   : %08x\n", vfmt.fmt.pix.priv);
 		break;
+	case V4L2_BUF_TYPE_VIDEO_OUTPUT_OVERLAY:
 	case V4L2_BUF_TYPE_VIDEO_OVERLAY:
-		printf("\tType        : %s\n", buftype2s(vfmt.type).c_str());
 		printf("\tLeft/Top    : %d/%d\n",
 				vfmt.fmt.win.w.left, vfmt.fmt.win.w.top);
 		printf("\tWidth/Height: %d/%d\n",
 				vfmt.fmt.win.w.width, vfmt.fmt.win.w.height);
 		printf("\tField       : %s\n", field2s(vfmt.fmt.win.field).c_str());
-		// TODO: check G_FBUF capabilities
-		printf("\tChroma Key  : %08x\n", vfmt.fmt.win.chromakey);
+		printf("\tChroma Key  : 0x%08x\n", vfmt.fmt.win.chromakey);
+		printf("\tGlobal Alpha: 0x%02x\n", vfmt.fmt.win.global_alpha);
 		printf("\tClip Count  : %u\n", vfmt.fmt.win.clipcount);
+		printf("\tClip Bitmap : %s\n", vfmt.fmt.win.bitmap ? "Yes" : "No");
 		break;
 	case V4L2_BUF_TYPE_VBI_CAPTURE:
 	case V4L2_BUF_TYPE_VBI_OUTPUT:
-		printf("\tType            : %s\n", buftype2s(vfmt.type).c_str());
 		printf("\tSampling Rate   : %u Hz\n", vfmt.fmt.vbi.sampling_rate);
 		printf("\tOffset          : %u samples (%g secs after leading edge)\n",
 				vfmt.fmt.vbi.offset,
@@ -632,7 +771,6 @@ static int printfmt(struct v4l2_format vfmt)
 		break;
 	case V4L2_BUF_TYPE_SLICED_VBI_CAPTURE:
 	case V4L2_BUF_TYPE_SLICED_VBI_OUTPUT:
-		printf("\tType           : %s\n", buftype2s(vfmt.type).c_str());
 		printf("\tService Set    : %s\n",
 				flags2s(vfmt.fmt.sliced.service_set, service_def).c_str());
 		for (int i = 0; i < 24; i++) {
@@ -643,10 +781,8 @@ static int printfmt(struct v4l2_format vfmt)
 		printf("\tI/O Size       : %u\n", vfmt.fmt.sliced.io_size);
 		break;
 	case V4L2_BUF_TYPE_PRIVATE:
-		printf("\tType: %s\n", buftype2s(vfmt.type).c_str());
 		break;
 	default:
-		printf("\tType: %s\n", buftype2s(vfmt.type).c_str());
 		return -1;
 	}
 	return 0;
@@ -752,6 +888,8 @@ static std::string cap2s(unsigned cap)
 		s += "\t\tVideo Output\n";
 	if (cap & V4L2_CAP_VIDEO_OVERLAY)
 		s += "\t\tVideo Overlay\n";
+	if (cap & V4L2_CAP_VIDEO_OUTPUT_OVERLAY)
+		s += "\t\tVideo Output Overlay\n";
 	if (cap & V4L2_CAP_VBI_CAPTURE)
 		s += "\t\tVBI Capture\n";
 	if (cap & V4L2_CAP_VBI_OUTPUT)
@@ -931,6 +1069,62 @@ static void print_std(const char *prefix, const char *stds[], unsigned long long
 	printf("\n");
 }
 
+static void do_crop(int fd, unsigned int set_crop, struct v4l2_rect &vcrop, v4l2_buf_type type)
+{
+    struct v4l2_crop in_crop;
+
+    in_crop.type = type;
+    if (ioctl(fd, VIDIOC_G_CROP, &in_crop) < 0)
+	fprintf(stderr, "ioctl: VIDIOC_G_CROP failed\n");
+    else {
+	if (set_crop & CropWidth)
+	    in_crop.c.width = vcrop.width;
+	if (set_crop & CropHeight)
+	    in_crop.c.height = vcrop.height;
+	if (set_crop & CropLeft)
+	    in_crop.c.left = vcrop.left;
+	if (set_crop & CropTop)
+	    in_crop.c.top = vcrop.top;
+	if (ioctl(fd, VIDIOC_S_CROP, &in_crop) < 0)
+	    fprintf(stderr, "ioctl: VIDIOC_S_CROP failed\n");
+    }
+}
+
+static void parse_crop(char *optarg, unsigned int &set_crop, v4l2_rect &vcrop)
+{
+    char *value;
+    char *subs = optarg;
+
+    while (*subs != '\0') {
+	static char *const subopts[] = {
+	    "left",
+	    "top",
+	    "width",
+	    "height",
+	    NULL
+	};
+
+	switch (parse_subopt(&subs, subopts, &value)) {
+	    case 0:
+		vcrop.left = strtol(value, 0L, 0);
+		set_crop |= CropLeft;
+		break;
+	    case 1:
+		vcrop.top = strtol(value, 0L, 0);
+		set_crop |= CropTop;
+		break;
+	    case 2:
+		vcrop.width = strtol(value, 0L, 0);
+		set_crop |= CropWidth;
+		break;
+	    case 3:
+		vcrop.height = strtol(value, 0L, 0);
+		set_crop |= CropHeight;
+		break;
+	}
+    }
+}
+
 int main(int argc, char **argv)
 {
 	char *value, *subs;
@@ -942,6 +1136,11 @@ int main(int argc, char **argv)
 	unsigned int set_fmts = 0;
 	unsigned int set_fmts_out = 0;
 	unsigned int set_crop = 0;
+	unsigned int set_crop_out = 0;
+	unsigned int set_crop_overlay = 0;
+	unsigned int set_crop_out_overlay = 0;
+	unsigned int set_fbuf = 0;
+	unsigned int set_overlay_fmt_out = 0;
 
 	int mode = V4L2_TUNER_MODE_STEREO;	/* set audio mode */
 
@@ -954,6 +1153,7 @@ int main(int argc, char **argv)
 	struct v4l2_format vbi_fmt_out;	/* set_format/get_format for sliced VBI output */
 	struct v4l2_format raw_fmt;	/* set_format/get_format for VBI */
 	struct v4l2_format raw_fmt_out;	/* set_format/get_format for VBI output */
+	struct v4l2_format overlay_fmt_out;	/* set_format/get_format video overlay output */
 	struct v4l2_tuner tuner;        /* set_tuner/get_tuner */
 	struct v4l2_capability vcap;	/* list_cap */
 	struct v4l2_input vin;		/* list_inputs */
@@ -961,6 +1161,10 @@ int main(int argc, char **argv)
 	struct v4l2_audio vaudio;	/* list audio inputs */
 	struct v4l2_audioout vaudout;   /* audio outputs */
 	struct v4l2_rect vcrop; 	/* crop rect */
+	struct v4l2_rect vcrop_out; 	/* crop rect */
+	struct v4l2_rect vcrop_overlay; 	/* crop rect */
+	struct v4l2_rect vcrop_out_overlay; 	/* crop rect */
+	struct v4l2_framebuffer fbuf;   /* fbuf */
 	int input;			/* set_input/get_input */
 	int output;			/* set_output/get_output */
 	v4l2_std_id std;		/* get_std/set_std */
@@ -975,6 +1179,7 @@ int main(int argc, char **argv)
 	memset(&raw_fmt, 0, sizeof(raw_fmt));
 	memset(&vfmt_out, 0, sizeof(vfmt_out));
 	memset(&vbi_fmt_out, 0, sizeof(vbi_fmt_out));
+	memset(&overlay_fmt_out, 0, sizeof(overlay_fmt_out));
 	memset(&raw_fmt_out, 0, sizeof(raw_fmt_out));
 	memset(&tuner, 0, sizeof(tuner));
 	memset(&vcap, 0, sizeof(vcap));
@@ -983,8 +1188,12 @@ int main(int argc, char **argv)
 	memset(&vaudio, 0, sizeof(vaudio));
 	memset(&vaudout, 0, sizeof(vaudout));
 	memset(&vcrop, 0, sizeof(vcrop));
+	memset(&vcrop_out, 0, sizeof(vcrop_out));
+	memset(&vcrop_overlay, 0, sizeof(vcrop_overlay));
+	memset(&vcrop_out_overlay, 0, sizeof(vcrop_out_overlay));
 	memset(&vf, 0, sizeof(vf));
 	memset(&vs, 0, sizeof(vs));
+	memset(&fbuf, 0, sizeof(fbuf));
 
 	if (argc == 1) {
 		usage();
@@ -1031,11 +1240,11 @@ int main(int argc, char **argv)
 				switch (parse_subopt(&subs, subopts, &value)) {
 				case 0:
 					vfmt.fmt.pix.width = strtol(value, 0L, 0);
-					set_fmts |= FMTWidth;
+					set_fmts |= FmtWidth;
 					break;
 				case 1:
 					vfmt.fmt.pix.height = strtol(value, 0L, 0);
-					set_fmts |= FMTHeight;
+					set_fmts |= FmtHeight;
 					break;
 				}
 			}
@@ -1052,46 +1261,84 @@ int main(int argc, char **argv)
 				switch (parse_subopt(&subs, subopts, &value)) {
 				case 0:
 					vfmt_out.fmt.pix.width = strtol(value, 0L, 0);
-					set_fmts_out |= FMTWidth;
+					set_fmts_out |= FmtWidth;
 					break;
 				case 1:
 					vfmt_out.fmt.pix.height = strtol(value, 0L, 0);
-					set_fmts_out |= FMTHeight;
+					set_fmts_out |= FmtHeight;
 					break;
 				}
 			}
 			break;
-		case OptSetVideoCrop:
+		case OptSetOutputOverlayFormat:
 			subs = optarg;
 			while (*subs != '\0') {
 				static char *const subopts[] = {
-					"left",
-					"top",
-					"width",
-					"height",
+					"chromakey",
+					"global_alpha",
 					NULL
 				};
 
 				switch (parse_subopt(&subs, subopts, &value)) {
 				case 0:
-					vcrop.left = strtol(value, 0L, 0);
-					set_crop |= CropLeft;
+					overlay_fmt_out.fmt.win.chromakey = strtol(value, 0L, 0);
+					set_overlay_fmt_out |= FmtChromaKey;
 					break;
 				case 1:
-					vcrop.top = strtol(value, 0L, 0);
-					set_crop |= CropTop;
-					break;
-				case 2:
-					vcrop.width = strtol(value, 0L, 0);
-					set_crop |= CropWidth;
-					break;
-				case 3:
-					vcrop.height = strtol(value, 0L, 0);
-					set_crop |= CropHeight;
+					overlay_fmt_out.fmt.win.global_alpha = strtol(value, 0L, 0);
+					set_overlay_fmt_out |= FmtGlobalAlpha;
 					break;
 				}
 			}
 			break;
+		case OptSetFBuf:
+			subs = optarg;
+			while (*subs != '\0') {
+				static char *const subopts[] = {
+					"chromakey",
+					"global_alpha",
+					"local_alpha",
+					"global_inv_alpha",
+					"local_inv_alpha",
+					NULL
+				};
+
+				switch (parse_subopt(&subs, subopts, &value)) {
+				case 0:
+					fbuf.flags |= strtol(value, 0L, 0) ? V4L2_FBUF_FLAG_CHROMAKEY : 0;
+					set_fbuf |= V4L2_FBUF_FLAG_CHROMAKEY;
+					break;
+				case 1:
+					fbuf.flags |= strtol(value, 0L, 0) ? V4L2_FBUF_FLAG_GLOBAL_ALPHA : 0;
+					set_fbuf |= V4L2_FBUF_FLAG_GLOBAL_ALPHA;
+					break;
+				case 2:
+					fbuf.flags |= strtol(value, 0L, 0) ? V4L2_FBUF_FLAG_LOCAL_ALPHA : 0;
+					set_fbuf |= V4L2_FBUF_FLAG_LOCAL_ALPHA;
+					break;
+				case 3:
+					fbuf.flags |= strtol(value, 0L, 0) ? V4L2_FBUF_FLAG_GLOBAL_INV_ALPHA : 0;
+					set_fbuf |= V4L2_FBUF_FLAG_GLOBAL_INV_ALPHA;
+					break;
+				case 4:
+					fbuf.flags |= strtol(value, 0L, 0) ? V4L2_FBUF_FLAG_LOCAL_INV_ALPHA : 0;
+					set_fbuf |= V4L2_FBUF_FLAG_LOCAL_INV_ALPHA;
+					break;
+				}
+			}
+			break;
+		case OptSetCrop:
+			parse_crop(optarg, set_crop, vcrop);
+			break;
+		case OptSetOutputCrop:
+			parse_crop(optarg, set_crop_out, vcrop_out);
+			break;
+		case OptSetOverlayCrop:
+			parse_crop(optarg, set_crop_overlay, vcrop_overlay);
+			break;
+		case OptSetOutputOverlayCrop:
+			parse_crop(optarg, set_crop_out_overlay, vcrop_out_overlay);
+			break;
 		case OptSetInput:
 			input = strtol(optarg, 0L, 0);
 			break;
@@ -1251,6 +1498,7 @@ int main(int argc, char **argv)
 	free(device);
 
 	doioctl(fd, VIDIOC_QUERYCAP, &vcap, "VIDIOC_QUERYCAP");
+	capabilities = vcap.capabilities;
 	find_controls(fd);
 	for (ctrl_get_list::iterator iter = get_ctrls.begin(); iter != get_ctrls.end(); ++iter) {
 	    if (ctrl_str2id.find(*iter) == ctrl_str2id.end()) {
@@ -1267,7 +1515,7 @@ int main(int argc, char **argv)
 
 	if (options[OptAll]) {
 		options[OptGetVideoFormat] = 1;
-		options[OptGetVideoCrop] = 1;
+		options[OptGetCrop] = 1;
 		options[OptGetVideoOutFormat] = 1;
 		options[OptGetDriverInfo] = 1;
 		options[OptGetInput] = 1;
@@ -1278,16 +1526,20 @@ int main(int argc, char **argv)
 		options[OptGetFreq] = 1;
 		options[OptGetTuner] = 1;
 		options[OptGetOverlayFormat] = 1;
+		options[OptGetOutputOverlayFormat] = 1;
 		options[OptGetVbiFormat] = 1;
 		options[OptGetVbiOutFormat] = 1;
 		options[OptGetSlicedVbiFormat] = 1;
 		options[OptGetSlicedVbiOutFormat] = 1;
+		options[OptGetFBuf] = 1;
+		options[OptGetCropCap] = 1;
+		options[OptGetOutputCropCap] = 1;
 	}
 
 	/* Information Opts */
 
 	if (options[OptGetDriverInfo]) {
-		printf("Driver info:\n");
+		printf("Driver Info:\n");
 		printf("\tDriver name   : %s\n", vcap.driver);
 		printf("\tCard type     : %s\n", vcap.card);
 		printf("\tBus info      : %s\n", vcap.bus_info);
@@ -1373,9 +1625,9 @@ int main(int argc, char **argv)
 		if (ioctl(fd, VIDIOC_G_FMT, &in_vfmt) < 0)
 			fprintf(stderr, "ioctl: VIDIOC_G_FMT failed\n");
 		else {
-			if (set_fmts & FMTWidth)
+			if (set_fmts & FmtWidth)
 				in_vfmt.fmt.pix.width = vfmt.fmt.pix.width;
-			if (set_fmts & FMTHeight)
+			if (set_fmts & FmtHeight)
 				in_vfmt.fmt.pix.height = vfmt.fmt.pix.height;
 			if (ioctl(fd, VIDIOC_S_FMT, &in_vfmt) < 0)
 				fprintf(stderr, "ioctl: VIDIOC_S_FMT failed\n");
@@ -1389,10 +1641,10 @@ int main(int argc, char **argv)
 		if (ioctl(fd, VIDIOC_G_FMT, &in_vfmt) < 0)
 			fprintf(stderr, "ioctl: VIDIOC_G_FMT failed\n");
 		else {
-			if (set_fmts & FMTWidth)
-				in_vfmt.fmt.pix.width = vfmt.fmt.pix.width;
-			if (set_fmts & FMTHeight)
-				in_vfmt.fmt.pix.height = vfmt.fmt.pix.height;
+			if (set_fmts_out & FmtWidth)
+				in_vfmt.fmt.pix.width = vfmt_out.fmt.pix.width;
+			if (set_fmts_out & FmtHeight)
+				in_vfmt.fmt.pix.height = vfmt_out.fmt.pix.height;
 			if (ioctl(fd, VIDIOC_S_FMT, &in_vfmt) < 0)
 				fprintf(stderr, "ioctl: VIDIOC_S_FMT failed\n");
 		}
@@ -1422,26 +1674,50 @@ int main(int argc, char **argv)
 		}
 	}
 
-	if (options[OptSetVideoCrop]) {
-		struct v4l2_crop in_crop;
-		printf("ioctl: VIDIOC_S_CROP\n");
-		in_crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
-		if (ioctl(fd, VIDIOC_G_CROP, &in_crop) < 0)
-			fprintf(stderr, "ioctl: VIDIOC_G_CROP failed\n");
+	if (options[OptSetOutputOverlayFormat]) {
+		struct v4l2_format fmt;
+		memset(&fmt, 0, sizeof(fmt));
+		fmt.type = V4L2_BUF_TYPE_VIDEO_OUTPUT_OVERLAY;
+		if (ioctl(fd, VIDIOC_G_FMT, &fmt, "VIDIOC_G_FMT") < 0)
+			fprintf(stderr, "ioctl: VIDIOC_G_FMT failed\n");
 		else {
-			if (set_crop & CropWidth)
-				in_crop.c.width = vcrop.width;
-			if (set_crop & CropHeight)
-				in_crop.c.height = vcrop.height;
-			if (set_crop & CropLeft)
-				in_crop.c.left = vcrop.left;
-			if (set_crop & CropTop)
-				in_crop.c.top = vcrop.top;
-			if (ioctl(fd, VIDIOC_S_CROP, &in_crop) < 0)
-				fprintf(stderr, "ioctl: VIDIOC_S_CROP failed\n");
+			if (set_overlay_fmt_out & FmtChromaKey)
+				fmt.fmt.win.chromakey = overlay_fmt_out.fmt.win.chromakey;
+			if (set_overlay_fmt_out & FmtGlobalAlpha)
+				fmt.fmt.win.global_alpha = overlay_fmt_out.fmt.win.global_alpha;
+			if (ioctl(fd, VIDIOC_S_FMT, &fmt, "VIDIOC_S_FMT") < 0)
+				fprintf(stderr, "ioctl: VIDIOC_S_FMT failed\n");
 		}
 	}
 
+	if (options[OptSetFBuf]) {
+		struct v4l2_framebuffer fb;
+		if (ioctl(fd, VIDIOC_G_FBUF, &fb, "VIDIOC_G_FBUF") < 0)
+			fprintf(stderr, "ioctl: VIDIOC_G_FBUF failed\n");
+		else {
+			fb.flags &= ~set_fbuf;
+			fb.flags |= fbuf.flags;
+			if (ioctl(fd, VIDIOC_S_FBUF, &fb) < 0)
+				fprintf(stderr, "ioctl: VIDIOC_S_FBUF failed\n");
+		}
+	}
+
+	if (options[OptSetCrop]) {
+		do_crop(fd, set_crop, vcrop, V4L2_BUF_TYPE_VIDEO_CAPTURE);
+	}
+
+	if (options[OptSetOutputCrop]) {
+		do_crop(fd, set_crop_out, vcrop_out, V4L2_BUF_TYPE_VIDEO_OUTPUT);
+	}
+
+	if (options[OptSetOverlayCrop]) {
+		do_crop(fd, set_crop_overlay, vcrop_overlay, V4L2_BUF_TYPE_VIDEO_OVERLAY);
+	}
+
+	if (options[OptSetOutputOverlayCrop]) {
+		do_crop(fd, set_crop_out_overlay, vcrop_out_overlay, V4L2_BUF_TYPE_VIDEO_OUTPUT_OVERLAY);
+	}
+
 	if (options[OptSetCtrl] && !set_ctrls.empty()) {
 		struct v4l2_ext_controls ctrls = { 0 };
 
@@ -1510,6 +1786,15 @@ int main(int argc, char **argv)
 				fprintf(stderr, "error printing result\n");
 	}
 
+	if (options[OptGetOutputOverlayFormat]) {
+		struct v4l2_format fmt;
+		memset(&fmt, 0, sizeof(fmt));
+		fmt.type = V4L2_BUF_TYPE_VIDEO_OUTPUT_OVERLAY;
+		if (doioctl(fd, VIDIOC_G_FMT, &fmt, "VIDIOC_G_FMT") == 0)
+			if (printfmt(fmt) != 0)
+				fprintf(stderr, "error printing result\n");
+	}
+
 	if (options[OptGetSlicedVbiFormat]) {
 		vbi_fmt.type = V4L2_BUF_TYPE_SLICED_VBI_CAPTURE;
 		if (doioctl(fd, VIDIOC_G_FMT, &vbi_fmt,
@@ -1546,7 +1831,45 @@ int main(int argc, char **argv)
 					"error printing result\n");
 	}
 
-	if (options[OptGetVideoCrop]) {
+	if (options[OptGetFBuf]) {
+		struct v4l2_framebuffer fb;
+		if (doioctl(fd, VIDIOC_G_FBUF, &fb, "VIDIOC_G_FBUF") == 0)
+			printfbuf(fb);
+	}
+
+	if (options[OptGetCropCap]) {
+		struct v4l2_cropcap cropcap;
+
+		cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
+		if (doioctl(fd, VIDIOC_CROPCAP, &cropcap, "VIDIOC_CROPCAP") == 0)
+			printcropcap(cropcap);
+	}
+
+	if (options[OptGetOutputCropCap]) {
+		struct v4l2_cropcap cropcap;
+
+		cropcap.type = V4L2_BUF_TYPE_VIDEO_OUTPUT;
+		if (doioctl(fd, VIDIOC_CROPCAP, &cropcap, "VIDIOC_CROPCAP") == 0)
+			printcropcap(cropcap);
+	}
+
+	if (options[OptGetOverlayCropCap]) {
+		struct v4l2_cropcap cropcap;
+
+		cropcap.type = V4L2_BUF_TYPE_VIDEO_OVERLAY;
+		if (doioctl(fd, VIDIOC_CROPCAP, &cropcap, "VIDIOC_CROPCAP") == 0)
+			printcropcap(cropcap);
+	}
+
+	if (options[OptGetOutputOverlayCropCap]) {
+		struct v4l2_cropcap cropcap;
+
+		cropcap.type = V4L2_BUF_TYPE_VIDEO_OUTPUT_OVERLAY;
+		if (doioctl(fd, VIDIOC_CROPCAP, &cropcap, "VIDIOC_CROPCAP") == 0)
+			printcropcap(cropcap);
+	}
+
+	if (options[OptGetCrop]) {
 		struct v4l2_crop crop;
 
 		crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
@@ -1554,6 +1877,30 @@ int main(int argc, char **argv)
 			printcrop(crop);
 	}
 
+	if (options[OptGetOutputCrop]) {
+		struct v4l2_crop crop;
+
+		crop.type = V4L2_BUF_TYPE_VIDEO_OUTPUT;
+		if (doioctl(fd, VIDIOC_G_CROP, &crop, "VIDIOC_G_CROP") == 0)
+			printcrop(crop);
+	}
+
+	if (options[OptGetOverlayCrop]) {
+		struct v4l2_crop crop;
+
+		crop.type = V4L2_BUF_TYPE_VIDEO_OVERLAY;
+		if (doioctl(fd, VIDIOC_G_CROP, &crop, "VIDIOC_G_CROP") == 0)
+			printcrop(crop);
+	}
+
+	if (options[OptGetOutputOverlayCrop]) {
+		struct v4l2_crop crop;
+
+		crop.type = V4L2_BUF_TYPE_VIDEO_OUTPUT_OVERLAY;
+		if (doioctl(fd, VIDIOC_G_CROP, &crop, "VIDIOC_G_CROP") == 0)
+			printcrop(crop);
+	}
+
 	if (options[OptGetInput]) {
 		if (doioctl(fd, VIDIOC_G_INPUT, &input, "VIDIOC_G_INPUT") == 0) {
 			printf("Video input : %d", input);
@@ -1617,7 +1964,7 @@ int main(int argc, char **argv)
 				NULL
 			};
 
-			printf("Video standard = 0x%08llx\n", (unsigned long long)std);
+			printf("Video Standard = 0x%08llx\n", (unsigned long long)std);
 			if (std & 0xfff) {
 				print_std("PAL", pal, std);
 			}
@@ -1815,7 +2162,7 @@ int main(int argc, char **argv)
 	if (options[OptGetSlicedVbiCap]) {
 		struct v4l2_sliced_vbi_cap cap;
 
-		//cap.type = V4L2_BUF_TYPE_VBI_CAPTURE;
+		cap.type = V4L2_BUF_TYPE_VBI_CAPTURE;
 		if (doioctl(fd, VIDIOC_G_SLICED_VBI_CAP, &cap, "VIDIOC_G_SLICED_VBI_CAP") == 0) {
 			print_sliced_vbi_cap(cap);
 		}
@@ -1824,7 +2171,7 @@ int main(int argc, char **argv)
 	if (options[OptGetSlicedVbiOutCap]) {
 		struct v4l2_sliced_vbi_cap cap;
 
-		//cap.type = V4L2_BUF_TYPE_VBI_OUTPUT;
+		cap.type = V4L2_BUF_TYPE_VBI_OUTPUT;
 		if (doioctl(fd, VIDIOC_G_SLICED_VBI_CAP, &cap, "VIDIOC_G_SLICED_VBI_CAP") == 0) {
 			print_sliced_vbi_cap(cap);
 		}
